#Exercise 3 template for CHE210D

#import python modules
import numpy as np
import time
import matplotlib.pyplot as plt



import sys
import os
sys.path.append(r'/home/vikram/COURSES/210D/LJFluidFE_corrected')
#import compiled fortran library
import ex4lib

#set the random number seed; useful for debugging
##np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [ex4lib.calcenergy(Pos + NormDir * x, M, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = ex4lib.calcenergy(Pos + NormDir * dx, M, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [ex4lib.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [ex4lib.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print("Warning: maximum number of iterations reached in line search.")
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = ex4lib.calcenergy(Pos + NormDir * d3, M, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = ex4lib.calcenergy(Pos + NormDir * d3, M, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = ex4lib.calcenergyforces(Pos, M, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = ex4lib.calcenergyforces(Pos, M, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos

#define ProposalMove():
    
#### YOUR CODE HERE ####

#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####
    
M=1

N_poly=int(240/M)
N=N_poly*M

rho=0.8

L = np.power(N/rho,1./3.)

#distance cutoff for pairwise interactions
Cut = 2.5

T = 3.0




InitCoords=InitPositions(N, L) #Initialize coords 
InitPE=ex4lib.calcenergy(InitCoords, M, L, Cut) #compute Initial PE
print("Initial PE before minimization is: ", InitPE) 

dx=0.001
EFracTolLS = 1.e-8
EFracTolCG = 1.e-10
PEMin, Pos = ConjugateGradient(InitCoords, dx, EFracTolLS, EFracTolCG) #energy min
print("PE after minimization is: ", PEMin) 

def getdeltaUs(lmbCtr,Ui,Pos):
    
    global M,L,Cut
    global lmbLJArr,nLmb
    
    if lmbCtr==(nLmb-1):
        U_F=float('inf')
    else:
        U_F=ex4lib.lmbdaljcalcenergy(lmbLJArr[lmbCtr+1], Pos, M, L, Cut)[0]-Ui

    if lmbCtr==0:
        U_R=float('-inf')
    else:
        U_R=Ui-ex4lib.lmbdaljcalcenergy(lmbLJArr[lmbCtr-1], Pos, M, L, Cut)[0]
    return U_F,U_R

def qgetdeltaUs(q,lmbCtr,Ui,Pos):
    
    global M,L,Cut
    global lmbLJArr,nLmb
    
    if lmbCtr==(nLmb-1):
        U_F=float('inf')
    else:
        U_F=ex4lib.qlmbdaljcalcenergy(q,lmbLJArr[lmbCtr+1], Pos, M, L, Cut)[0]-Ui

    if lmbCtr==0:
        U_R=float('-inf')
    else:
        U_R=Ui-ex4lib.qlmbdaljcalcenergy(q,lmbLJArr[lmbCtr-1], Pos, M, L, Cut)[0]
    return U_F,U_R
    
def MCRun(lmbCtr, tstepMax, drMax, FileName, Pos):
    
    global M, L, Cut

    global lmbLJArr,nLmb
    #current lmb value
    lmbLJ=lmbLJArr[lmbCtr]

    tstep=0 #MCStep start ctr
    
    Forces = np.zeros_like(Pos) #just to avoid redifing(deleting) Forces from arguments everywhere in old script
    U, dUdlmb= ex4lib.lmbdaljcalcenergy(lmbLJ, Pos, M, L, Cut)
    
    U_F,U_R=getdeltaUs(lmbCtr,U,Pos)

            
    
    
    tAcc=0
    dUPosAcc=0

    fOutput=open(FileName,'w')
    fOutput.write("%12s %16s %16s %16s %16s\n"%("Step", "PotentialEnergy", "dUdlmb", "U_F", "U_R"))
    fOutput.write("%12d %16.8f %16.8f %16.8f %16.8f\n"%(tstep, U, dUdlmb, U_F, U_R))       
    
    while tstep < tstepMax:

        tstep = tstep + 1
        accept=False
        MoveAtom = np.random.randint(N)
        dr=np.random.uniform(-drMax,drMax,3)
#        OldPosAll=Pos.copy()
        OldPos = Pos[MoveAtom,:].copy()
        U0,dUdlmb0=ex4lib.qlmbdaljcalcenergy(MoveAtom, lmbLJ, Pos, M, L, Cut)
        U_F0,U_R0=qgetdeltaUs(MoveAtom,lmbCtr,U0,Pos)
        Pos[MoveAtom,:]=OldPos + dr
        U1,dUdlmb1=ex4lib.qlmbdaljcalcenergy(MoveAtom, lmbLJ, Pos, M, L, Cut)
        U_F1,U_R1=qgetdeltaUs(MoveAtom,lmbCtr,U1,Pos)        
    
        dU=U1-U0
        DdUdlmb=dUdlmb1-dUdlmb0
        
        dU_F=U_F1-U_F0
        dU_R=U_R1-U_R0
        
        if dU<0:
            accept=True
        else:
            Pacc=np.exp(-dU/T)
            randNum=np.random.uniform()
            if Pacc>randNum:
                dUPosAcc=dUPosAcc+1
                accept=True
        
        if accept:
            U = U + dU
            dUdlmb = dUdlmb + DdUdlmb
            U_F=U_F+dU_F
            U_R=U_R+dU_R
            tAcc=tAcc+1
        else:
            Pos[MoveAtom,:]=OldPos
        
        if tstep%RefreshRate == 0:
            U, dUdlmb = ex4lib.lmbdaljcalcenergy(lmbLJ, Pos, M, L, Cut)
            U_F,U_R=getdeltaUs(lmbCtr,U,Pos)

        if tstep%DumpRate == 0:
            fOutput.write("%12d %16.8f %16.8f %16.8f %16.8f\n"%(tstep, U, dUdlmb, U_F, U_R))   
            


    #End of while loop

    fOutput.close()    
    
    AccRatio=tAcc*1.0/tstepMax
    
    print(AccRatio)
    

    FileNameSettings=r'Settings'+FileName
    fOutput=open(FileNameSettings,'w')
    fOutput.write("%-12s %16.8f \n"%("lmbLJ=",lmbLJ))
    fOutput.write("%-12s %16.8f \n"%("lmbCtr=",lmbCtr))
    fOutput.write("%-12s %16.8f \n"%("M=",M))
    fOutput.write("%-12s %16.8f \n"%("T=",T)) 
    fOutput.write("%-12s %16.8f \n"%("rho=",rho))
    fOutput.write("%-12s %16.8f \n"%("NPoly=",N_poly))
    fOutput.write("%-12s %16.8f \n"%("N=",N))
    fOutput.write("%-12s %16.8f \n"%("L=",L))
    fOutput.write("%-12s %16d \n"%("MCSteps=",tstepMax))
    fOutput.write("%-12s %16.8f \n"%("drMax=",drMax))  
    fOutput.write("%-12s %16.8f \n"%("AccRatio=",AccRatio)) 
    fOutput.close()
    
    return tstep, Pos, U, dUdlmb, AccRatio


tstepMax=300*1000*N

lmbLJArr=np.array([0.00, 6.98177409e-04, 3.67500078e-03, 9.01564248e-03, 1.66958448e-02,
       2.66791545e-02, 3.89180316e-02, 5.33541641e-02, 6.99187620e-02,
       8.85328897e-02, 1.09107844e-01, 1.31545576e-01, 1.55739160e-01,
       1.81573303e-01, 2.08924894e-01, 2.37663590e-01, 2.67652438e-01,
       2.98748528e-01, 3.30803673e-01, 3.63665115e-01, 3.97176255e-01,
       4.31177397e-01, 4.65506510e-01, 5.00000000e-01, 5.34493490e-01,
       5.68822603e-01, 6.02823745e-01, 6.36334885e-01, 6.69196327e-01,
       7.01251472e-01, 7.32347562e-01, 7.62336410e-01, 7.91075106e-01,
       8.18426697e-01, 8.44260840e-01, 8.68454424e-01, 8.90892156e-01,
       9.11467110e-01, 9.30081238e-01, 9.46645836e-01, 9.61081968e-01,
       9.73320845e-01, 9.83304155e-01, 9.90984358e-01, 9.96324999e-01,
       9.99301823e-01, 1.00])
nLmb=len(lmbLJArr)    
    

lmbCtr = 11


drMax=0.15


RefreshRate=100*N


DumpRate=1*N


tstep, Pos, U, dUdlmb, AccRatio = MCRun(lmbCtr, tstepMax,drMax,r'EqmRun.txt',Pos)

##############################################################################################################################################